@@ -47,7 +47,7 @@
 
 require 'socket'
 require 'thread'
-require 'fcntl'
+require 'io/wait'
 require 'drb/eq'
 
 #
@@ -177,6 +177,9 @@ require 'drb/eq'
 #   # Not necessary for this small example, but will be required
 #   # as soon as we pass a non-marshallable object as an argument
 #   # to a dRuby call.
+#   #
+#   # Note: this must be called at least once per process to take any effect.
+#   # This is particularly important if your application forks.
 #   DRb.start_service
 #
 #   timeserver = DRbObject.new_with_uri(SERVER_URI)
@@ -231,7 +234,7 @@ require 'drb/eq'
 #       def get_logger(name)
 #           if !@loggers.has_key? name
 #               # make the filename safe, then declare it to be so
-#               fname = name.gsub(/[.\/]/, "_").untaint
+#               fname = name.gsub(/[.\/\\\:]/, "_").untaint
 #               @loggers[name] = Logger.new(name, @basedir + "/" + fname)
 #           end
 #           return @loggers[name]
@@ -426,6 +429,8 @@ module DRb
 
   # An exception wrapping an error object
   class DRbRemoteError < DRbError
+
+    # Creates a new remote error that wraps the Exception +error+
     def initialize(error)
       @reason = error.class.to_s
       super("#{error.message} (#{error.class})")
@@ -505,7 +510,16 @@ module DRb
     end
   end
 
+  # An Array wrapper that can be sent to another server via DRb.
+  #
+  # All entries in the array will be dumped or be references that point to
+  # the local server.
+
   class DRbArray
+
+    # Creates a new DRbArray that either dumps or wraps all the items in the
+    # Array +ary+ so they can be loaded by a remote DRb server.
+
     def initialize(ary)
       @ary = ary.collect { |obj|
         if obj.kind_of? DRbUndumped
@@ -521,11 +535,11 @@ module DRb
       }
     end
 
-    def self._load(s)
+    def self._load(s) # :nodoc:
       Marshal::load(s)
     end
 
-    def _dump(lv)
+    def _dump(lv) # :nodoc:
       Marshal.dump(@ary)
     end
   end
@@ -629,7 +643,7 @@ module DRb
     end
 
     private
-    def make_proxy(obj, error=false)
+    def make_proxy(obj, error=false) # :nodoc:
       if error
         DRbRemoteError.new(obj)
       else
@@ -731,7 +745,7 @@ module DRb
         end
       end
       if first && (config[:auto_load] != false)
-        auto_load(uri, config)
+        auto_load(uri)
         return open(uri, config, false)
       end
       raise DRbBadURI, 'can\'t parse uri:' + uri
@@ -755,7 +769,7 @@ module DRb
         end
       end
       if first && (config[:auto_load] != false)
-        auto_load(uri, config)
+        auto_load(uri)
         return open_server(uri, config, false)
       end
       raise DRbBadURI, 'can\'t parse uri:' + uri
@@ -778,14 +792,14 @@ module DRb
         end
       end
       if first && (config[:auto_load] != false)
-        auto_load(uri, config)
+        auto_load(uri)
         return uri_option(uri, config, false)
       end
       raise DRbBadURI, 'can\'t parse uri:' + uri
     end
     module_function :uri_option
 
-    def auto_load(uri, config)  # :nodoc:
+    def auto_load(uri)  # :nodoc:
       if uri =~ /^drb([a-z0-9]+):/
         require("drb/#{$1}") rescue nil
       end
@@ -793,10 +807,13 @@ module DRb
     module_function :auto_load
   end
 
-  # The default drb protocol.
+  # The default drb protocol which communicates over a TCP socket.
   #
-  # Communicates over a TCP socket.
+  # The DRb TCP protocol URI looks like:
+  # <code>druby://<host>:<port>?<option></code>.  The option is optional.
+
   class DRbTCPSocket
+    # :stopdoc:
     private
     def self.parse_uri(uri)
       if uri =~ /^druby:\/\/(.*?):(\d+)(\?(.*))?$/
@@ -812,7 +829,12 @@ module DRb
 
     public
 
-    # Open a client connection to +uri+ using configuration +config+.
+    # Open a client connection to +uri+ (DRb URI string) using configuration
+    # +config+.
+    #
+    # This can raise DRb::DRbBadScheme or DRb::DRbBadURI if +uri+ is not for a
+    # recognized protocol.  See DRb::DRbServer.new for information on built-in
+    # URI protocols.
     def self.open(uri, config)
       host, port, = parse_uri(uri)
       host.untaint
@@ -821,6 +843,7 @@ module DRb
       self.new(uri, soc, config)
     end
 
+    # Returns the hostname of this server
     def self.getservername
       host = Socket::gethostname
       begin
@@ -830,6 +853,9 @@ module DRb
       end
     end
 
+    # For the families available for +host+, returns a TCPServer on +port+.
+    # If +port+ is 0 the first available port is used.  IPv4 servers are
+    # preferred over IPv6 servers.
     def self.open_server_inaddr_any(host, port)
       infos = Socket::getaddrinfo(host, nil,
                                   Socket::AF_UNSPEC,
@@ -840,6 +866,7 @@ module DRb
       return TCPServer.open('0.0.0.0', port) if families.has_key?('AF_INET')
       return TCPServer.open('::', port) if families.has_key?('AF_INET6')
       return TCPServer.open(port)
+      # :stopdoc:
     end
 
     # Open a server listening for connections at +uri+ using
@@ -878,6 +905,7 @@ module DRb
       @acl = config[:tcp_acl]
       @msg = DRbMessage.new(config)
       set_sockopt(@socket)
+      @shutdown_pipe_r, @shutdown_pipe_w = IO.pipe
     end
 
     # Get the URI that we are connected to.
@@ -925,14 +953,28 @@ module DRb
         @socket.close
         @socket = nil
       end
+      close_shutdown_pipe
     end
 
+    def close_shutdown_pipe
+      if @shutdown_pipe_r && !@shutdown_pipe_r.closed?
+        @shutdown_pipe_r.close
+        @shutdown_pipe_r = nil
+      end
+      if @shutdown_pipe_w && !@shutdown_pipe_w.closed?
+        @shutdown_pipe_w.close
+        @shutdown_pipe_w = nil
+      end
+    end
+    private :close_shutdown_pipe
+
     # On the server side, for an instance returned by #open_server,
     # accept a client connection and return a new instance to handle
     # the server's side of this client-server session.
     def accept
       while true
-        s = @socket.accept
+        s = accept_or_shutdown
+        return nil unless s
         break if (@acl ? @acl.allow_socket?(s) : true)
         s.close
       end
@@ -944,10 +986,24 @@ module DRb
       self.class.new(uri, s, @config)
     end
 
+    def accept_or_shutdown
+      readables, = IO.select([@socket, @shutdown_pipe_r])
+      if readables.include? @shutdown_pipe_r
+        return nil
+      end
+      @socket.accept
+    end
+    private :accept_or_shutdown
+
+    # Graceful shutdown
+    def shutdown
+      @shutdown_pipe_w.close if @shutdown_pipe_w && !@shutdown_pipe_w.closed?
+    end
+
     # Check to see if this connection is alive.
     def alive?
       return false unless @socket
-      if IO.select([@socket], nil, nil, 0)
+      if @socket.to_io.wait_readable(0)
         close
         return false
       end
@@ -956,7 +1012,6 @@ module DRb
 
     def set_sockopt(soc) # :nodoc:
       soc.setsockopt(Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1)
-      soc.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC) if defined? Fcntl::FD_CLOEXEC
     end
   end
 
@@ -968,7 +1023,7 @@ module DRb
     def initialize(option)
       @option = option.to_s
     end
-    attr :option
+    attr_reader :option
     def to_s; @option; end
 
     def ==(other)
@@ -1008,10 +1063,13 @@ module DRb
       self.new_with(uri, ref)
     end
 
+    # Creates a DRb::DRbObject given the reference information to the remote
+    # host +uri+ and object +ref+.
+
     def self.new_with(uri, ref)
       it = self.allocate
-      it.instance_variable_set('@uri', uri)
-      it.instance_variable_set('@ref', ref)
+      it.instance_variable_set(:@uri, uri)
+      it.instance_variable_set(:@ref, ref)
       it
     end
 
@@ -1058,6 +1116,7 @@ module DRb
     undef :to_s
     undef :to_a if respond_to?(:to_a)
 
+    # Routes respond_to? to the referenced remote object.
     def respond_to?(msg_id, priv=false)
       case msg_id
       when :_dump
@@ -1069,7 +1128,7 @@ module DRb
       end
     end
 
-    # Routes method calls to the referenced object.
+    # Routes method calls to the referenced remote object.
     def method_missing(msg_id, *a, &b)
       if DRb.here?(@uri)
         obj = DRb.to_obj(@ref)
@@ -1094,7 +1153,8 @@ module DRb
       end
     end
 
-    def self.with_friend(uri)
+    # Given the +uri+ of another host executes the block provided.
+    def self.with_friend(uri) # :nodoc:
       friend = DRb.fetch_server(uri)
       return yield() unless friend
 
@@ -1105,7 +1165,9 @@ module DRb
       Thread.current['DRb'] = save if friend
     end
 
-    def self.prepare_backtrace(uri, result)
+    # Returns a modified backtrace from +result+ with the +uri+ where each call
+    # in the backtrace came from.
+    def self.prepare_backtrace(uri, result) # :nodoc:
       prefix = "(#{uri}) "
       bt = []
       result.backtrace.each do |x|
@@ -1236,9 +1298,9 @@ module DRb
       @@load_limit = sz
     end
 
-    # Set the default value for the :acl option.
+    # Set the default access control list to +acl+.  The default ACL is +nil+.
     #
-    # See #new().  The initial default value is nil.
+    # See also DRb::ACL and #new()
     def self.default_acl(acl)
       @@acl = acl
     end
@@ -1250,6 +1312,9 @@ module DRb
       @@idconv = idconv
     end
 
+    # Set the default safe level to +level+.  The default safe level is 0
+    #
+    # See #new for more information.
     def self.default_safe_level(level)
       @@safe_level = level
     end
@@ -1307,6 +1372,9 @@ module DRb
     # :argc_limit :: the maximum number of arguments to a remote
     #                method accepted by the server.  Defaults to
     #                256.
+    # :safe_level :: The safe level of the DRbServer.  The attribute
+    #                sets $SAFE for methods performed in the main_loop.
+    #                Defaults to 0.
     #
     # The default values of these options can be modified on
     # a class-wide basis by the class methods #default_argc_limit,
@@ -1366,6 +1434,10 @@ module DRb
     # The configuration of this DRbServer
     attr_reader :config
 
+    # The safe level for this server.  This is a number corresponding to
+    # $SAFE.
+    #
+    # The default safe_level is 0
     attr_reader :safe_level
 
     # Set whether to operate in verbose mode.
@@ -1383,6 +1455,7 @@ module DRb
       @thread.alive?
     end
 
+    # Is +uri+ the URI for this server?
     def here?(uri)
       @exported_uri.include?(uri)
     end
@@ -1393,7 +1466,12 @@ module DRb
       if  Thread.current['DRb'] && Thread.current['DRb']['server'] == self
         Thread.current['DRb']['stop_service'] = true
       else
-        @thread.kill.join
+        if @protocol.respond_to? :shutdown
+          @protocol.shutdown
+        else
+          [@thread, *@grp.list].each {|thread| thread.kill} # xxx: Thread#kill
+        end
+        @thread.join
       end
     end
 
@@ -1411,11 +1489,14 @@ module DRb
     end
 
     private
+
+    ##
+    # Starts the DRb main loop in a new thread.
+
     def run
       Thread.start do
         begin
-          while true
-            main_loop
+          while main_loop
           end
         ensure
           @protocol.close if @protocol
@@ -1545,16 +1626,9 @@ module DRb
 
     end
 
-    if RUBY_VERSION >= '1.8'
-      require 'drb/invokemethod'
-      class InvokeMethod
-        include InvokeMethod18Mixin
-      end
-    else
-      require 'drb/invokemethod16'
-      class InvokeMethod
-        include InvokeMethod16Mixin
-      end
+    require 'drb/invokemethod'
+    class InvokeMethod
+      include InvokeMethod18Mixin
     end
 
     # The main loop performed by a DRbServer's internal thread.
@@ -1565,7 +1639,9 @@ module DRb
     # returning responses, until the client closes the connection
     # or a local method call fails.
     def main_loop
-      Thread.start(@protocol.accept) do |client|
+      client0 = @protocol.accept
+      return nil if !client0
+      Thread.start(client0) do |client|
         @grp.add Thread.current
         Thread.current['DRb'] = { 'client' => client ,
                                   'server' => self }
@@ -1720,7 +1796,10 @@ module DRb
   end
   module_function :thread
 
-  # Set the default id conv object.
+  # Set the default id conversion object.
+  #
+  # This is expected to be an instance such as DRb::DRbIdConv that responds to
+  # #to_id and #to_obj that can convert objects to and from DRb references.
   #
   # See DRbServer#default_id_conv.
   def install_id_conv(idconv)
@@ -1728,7 +1807,7 @@ module DRb
   end
   module_function :install_id_conv
 
-  # Set the default acl.
+  # Set the default ACL to +acl+.
   #
   # See DRb::DRbServer.default_acl.
   def install_acl(acl)
@@ -1737,12 +1816,24 @@ module DRb
   module_function :install_acl
 
   @mutex = Mutex.new
-  def mutex
+  def mutex # :nodoc:
     @mutex
   end
   module_function :mutex
 
   @server = {}
+  # Registers +server+ with DRb.
+  #
+  # This is called when a new DRb::DRbServer is created.
+  #
+  # If there is no primary server then +server+ becomes the primary server.
+  #
+  # Example:
+  #
+  #  require 'drb'
+  #
+  #  s = DRb::DRbServer.new # automatically calls regist_server
+  #  DRb.fetch_server s.uri #=> #<DRb::DRbServer:0x...>
   def regist_server(server)
     @server[server.uri] = server
     mutex.synchronize do
@@ -1751,11 +1842,15 @@ module DRb
   end
   module_function :regist_server
 
+  # Removes +server+ from the list of registered servers.
   def remove_server(server)
     @server.delete(server.uri)
   end
   module_function :remove_server
 
+  # Retrieves the server with the given +uri+.
+  #
+  # See also regist_server and remove_server.
   def fetch_server(uri)
     @server[uri]
   end
