@@ -59,7 +59,6 @@ The original copyright notice follows.
    email: matumoto@math.keio.ac.jp
 */
 
-#include "ruby/ruby.h"
 #include "internal.h"
 
 #include <limits.h>
@@ -78,6 +77,12 @@ The original copyright notice follows.
 #include <sys/time.h>
 #endif
 
+#ifdef HAVE_SYSCALL_H
+#include <syscall.h>
+#elif defined HAVE_SYS_SYSCALL_H
+#include <sys/syscall.h>
+#endif
+
 #ifdef _WIN32
 # if !defined(_WIN32_WINNT) || _WIN32_WINNT < 0x0400
 #  undef _WIN32_WINNT
@@ -86,6 +91,7 @@ The original copyright notice follows.
 # endif
 #include <wincrypt.h>
 #endif
+#include "ruby_atomic.h"
 
 typedef int int_must_be_32bit_at_least[sizeof(int) * CHAR_BIT < 32 ? -1 : 1];
 
@@ -196,16 +202,26 @@ genrand_int32(struct MT *mt)
 }
 
 /* generates a random number on [0,1) with 53-bit resolution*/
+static double int_pair_to_real_exclusive(uint32_t a, uint32_t b);
 static double
 genrand_real(struct MT *mt)
 {
     /* mt must be initialized */
-    unsigned int a = genrand_int32(mt)>>5, b = genrand_int32(mt)>>6;
+    unsigned int a = genrand_int32(mt), b = genrand_int32(mt);
+    return int_pair_to_real_exclusive(a, b);
+}
+
+static double
+int_pair_to_real_exclusive(uint32_t a, uint32_t b)
+{
+    a >>= 5;
+    b >>= 6;
     return(a*67108864.0+b)*(1.0/9007199254740992.0);
 }
 
 /* generates a random number on [0,1] with 53-bit resolution*/
 static double int_pair_to_real_inclusive(uint32_t a, uint32_t b);
+#if 0
 static double
 genrand_real2(struct MT *mt)
 {
@@ -213,6 +229,7 @@ genrand_real2(struct MT *mt)
     uint32_t a = genrand_int32(mt), b = genrand_int32(mt);
     return int_pair_to_real_inclusive(a, b);
 }
+#endif
 
 /* These real versions are due to Isaku Wada, 2002/01/09 added */
 
@@ -335,6 +352,7 @@ static const rb_data_type_t random_data_type = {
 	random_free,
 	random_memsize,
     },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static rb_random_t *
@@ -376,18 +394,12 @@ rand_init(struct MT *mt, VALUE vseed)
     seed = rb_to_int(vseed);
 
     len = rb_absint_numwords(seed, 32, NULL);
-    if (MT_MAX_STATE < len)
-        len = MT_MAX_STATE;
     if (len > numberof(buf0))
         buf = ALLOC_N(unsigned int, len);
     sign = rb_integer_pack(seed, buf, len, sizeof(uint32_t), 0,
         INTEGER_PACK_LSWORD_FIRST|INTEGER_PACK_NATIVE_BYTE_ORDER);
     if (sign < 0)
         sign = -sign;
-    if (sign != 2) { /* not overflow */
-        while (0 < len && buf[len-1] == 0)
-            len--;
-    }
     if (len == 0) {
         buf[0] = 0;
         len = 1;
@@ -439,54 +451,130 @@ random_init(int argc, VALUE *argv, VALUE obj)
 # define USE_DEV_URANDOM 0
 #endif
 
+#if USE_DEV_URANDOM
+static int
+fill_random_bytes_urandom(void *seed, size_t size)
+{
+    /*
+      O_NONBLOCK and O_NOCTTY is meaningless if /dev/urandom correctly points
+      to a urandom device. But it protects from several strange hazard if
+      /dev/urandom is not a urandom device.
+    */
+    int fd = rb_cloexec_open("/dev/urandom",
+# ifdef O_NONBLOCK
+			     O_NONBLOCK|
+# endif
+# ifdef O_NOCTTY
+			     O_NOCTTY|
+# endif
+			     O_RDONLY, 0);
+    struct stat statbuf;
+    ssize_t ret = 0;
+
+    if (fd < 0) return -1;
+    rb_update_max_fd(fd);
+    if (fstat(fd, &statbuf) == 0 && S_ISCHR(statbuf.st_mode)) {
+	ret = read(fd, seed, size);
+    }
+    close(fd);
+    if (ret < 0 || (size_t)ret < size) return -1;
+    return 0;
+}
+#else
+# define fill_random_bytes_urandom(seed, size) -1
+#endif
+
+#if defined(_WIN32)
+static void
+release_crypt(void *p)
+{
+    HCRYPTPROV prov = (HCRYPTPROV)ATOMIC_PTR_EXCHANGE(*(HCRYPTPROV *)p, INVALID_HANDLE_VALUE);
+    if (prov && prov != (HCRYPTPROV)INVALID_HANDLE_VALUE) {
+	CryptReleaseContext(prov, 0);
+    }
+}
+
+static int
+fill_random_bytes_syscall(void *seed, size_t size)
+{
+    static HCRYPTPROV perm_prov;
+    HCRYPTPROV prov = perm_prov, old_prov;
+    if (!prov) {
+	if (!CryptAcquireContext(&prov, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) {
+	    prov = (HCRYPTPROV)INVALID_HANDLE_VALUE;
+	}
+	old_prov = (HCRYPTPROV)ATOMIC_PTR_CAS(perm_prov, 0, prov);
+	if (LIKELY(!old_prov)) { /* no other threads acquried */
+	    if (prov != (HCRYPTPROV)INVALID_HANDLE_VALUE) {
+		rb_gc_register_mark_object(Data_Wrap_Struct(0, 0, release_crypt, &perm_prov));
+	    }
+	}
+	else {			/* another thread acquried */
+	    if (prov != (HCRYPTPROV)INVALID_HANDLE_VALUE) {
+		CryptReleaseContext(prov, 0);
+	    }
+	    prov = old_prov;
+	}
+    }
+    if (prov == (HCRYPTPROV)INVALID_HANDLE_VALUE) return -1;
+    CryptGenRandom(prov, size, seed);
+    return 0;
+}
+#elif defined __linux__ && defined SYS_getrandom
+#include <linux/random.h>
+
+# ifndef GRND_NONBLOCK
+#   define GRND_NONBLOCK 0x0001	/* not defined in musl libc */
+# endif
+
+static int
+fill_random_bytes_syscall(void *seed, size_t size)
+{
+    static rb_atomic_t try_syscall = 1;
+    if (try_syscall) {
+	long ret;
+	errno = 0;
+	ret = syscall(SYS_getrandom, seed, size, GRND_NONBLOCK);
+	if (errno == ENOSYS) {
+	    ATOMIC_SET(try_syscall, 0);
+	    return -1;
+	}
+	if ((size_t)ret == size) return 0;
+    }
+    return -1;
+}
+#else
+# define fill_random_bytes_syscall(seed, size) -1
+#endif
+
+static int
+fill_random_bytes(void *seed, size_t size)
+{
+    int ret = fill_random_bytes_syscall(seed, size);
+    if (ret == 0) return ret;
+    return fill_random_bytes_urandom(seed, size);
+}
+
 static void
 fill_random_seed(uint32_t seed[DEFAULT_SEED_CNT])
 {
     static int n = 0;
     struct timeval tv;
-#if USE_DEV_URANDOM
-    int fd;
-    struct stat statbuf;
-#elif defined(_WIN32)
-    HCRYPTPROV prov;
-#endif
 
     memset(seed, 0, DEFAULT_SEED_LEN);
 
-#if USE_DEV_URANDOM
-    if ((fd = rb_cloexec_open("/dev/urandom", O_RDONLY
-#ifdef O_NONBLOCK
-            |O_NONBLOCK
-#endif
-#ifdef O_NOCTTY
-            |O_NOCTTY
-#endif
-            , 0)) >= 0) {
-        rb_update_max_fd(fd);
-        if (fstat(fd, &statbuf) == 0 && S_ISCHR(statbuf.st_mode)) {
-	    if (read(fd, seed, DEFAULT_SEED_LEN) < DEFAULT_SEED_LEN) {
-		/* abandon */;
-	    }
-        }
-        close(fd);
-    }
-#elif defined(_WIN32)
-    if (CryptAcquireContext(&prov, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) {
-	CryptGenRandom(prov, DEFAULT_SEED_LEN, (void *)seed);
-	CryptReleaseContext(prov, 0);
-    }
-#endif
+    fill_random_bytes(seed, sizeof(*seed));
 
     gettimeofday(&tv, 0);
     seed[0] ^= tv.tv_usec;
-    seed[1] ^= (unsigned int)tv.tv_sec;
+    seed[1] ^= (uint32_t)tv.tv_sec;
 #if SIZEOF_TIME_T > SIZEOF_INT
-    seed[0] ^= (unsigned int)((time_t)tv.tv_sec >> SIZEOF_INT * CHAR_BIT);
+    seed[0] ^= (uint32_t)((time_t)tv.tv_sec >> SIZEOF_INT * CHAR_BIT);
 #endif
     seed[2] ^= getpid() ^ (n++ << 16);
-    seed[3] ^= (unsigned int)(VALUE)&seed;
+    seed[3] ^= (uint32_t)(VALUE)&seed;
 #if SIZEOF_VOIDP > SIZEOF_INT
-    seed[2] ^= (unsigned int)((VALUE)&seed >> SIZEOF_INT * CHAR_BIT);
+    seed[2] ^= (uint32_t)((VALUE)&seed >> SIZEOF_INT * CHAR_BIT);
 #endif
 }
 
@@ -495,10 +583,10 @@ make_seed_value(const uint32_t *ptr)
 {
     VALUE seed;
     size_t len;
+    uint32_t buf[DEFAULT_SEED_CNT+1];
 
     if (ptr[DEFAULT_SEED_CNT-1] <= 1) {
         /* set leading-zero-guard */
-        uint32_t buf[DEFAULT_SEED_CNT+1];
         MEMCPY(buf, ptr, uint32_t, DEFAULT_SEED_CNT);
         buf[DEFAULT_SEED_CNT] = 1;
         ptr = buf;
@@ -530,6 +618,23 @@ random_seed(void)
     return make_seed_value(buf);
 }
 
+/*
+ * call-seq: Random.raw_seed(size) -> string
+ *
+ * Returns a raw seed string, using platform providing features.
+ *
+ *   Random.raw_seed(8)  #=> "\x78\x41\xBA\xAF\x7D\xEA\xD8\xEA"
+ */
+static VALUE
+random_raw_seed(VALUE self, VALUE size)
+{
+    long n = NUM2ULONG(size);
+    VALUE buf = rb_str_new(0, n);
+    if (n == 0) return buf;
+    if (fill_random_bytes(RSTRING_PTR(buf), n)) return Qnil;
+    return buf;
+}
+
 /*
  * call-seq: prng.seed -> integer
  *
@@ -632,7 +737,7 @@ random_load(VALUE obj, VALUE dump)
 
     rb_check_copyable(obj, dump);
     Check_Type(dump, T_ARRAY);
-    ary = RARRAY_RAWPTR(dump);
+    ary = RARRAY_CONST_PTR(dump);
     switch (RARRAY_LEN(dump)) {
       case 3:
 	seed = ary[2];
@@ -796,10 +901,9 @@ rb_genrand_ulong_limited(unsigned long limit)
     return limited_rand(default_mt(), limit);
 }
 
-unsigned int
-rb_random_int32(VALUE obj)
+static unsigned int
+random_int32(VALUE obj, rb_random_t *rnd)
 {
-    rb_random_t *rnd = try_get_rnd(obj);
     if (!rnd) {
 #if SIZEOF_LONG * CHAR_BIT > 32
 	VALUE lim = ULONG2NUM(0x100000000UL);
@@ -813,6 +917,25 @@ rb_random_int32(VALUE obj)
     return genrand_int32(&rnd->mt);
 }
 
+unsigned int
+rb_random_int32(VALUE obj)
+{
+    return random_int32(obj, try_get_rnd(obj));
+}
+
+static double
+random_real(VALUE obj, rb_random_t *rnd, int excl)
+{
+    uint32_t a = random_int32(obj, rnd);
+    uint32_t b = random_int32(obj, rnd);
+    if (excl) {
+	return int_pair_to_real_exclusive(a, b);
+    }
+    else {
+	return int_pair_to_real_inclusive(a, b);
+    }
+}
+
 double
 rb_random_real(VALUE obj)
 {
@@ -844,10 +967,9 @@ ulong_to_num_plus_1(unsigned long n)
 #endif
 }
 
-unsigned long
-rb_random_ulong_limited(VALUE obj, unsigned long limit)
+static unsigned long
+random_ulong_limited(VALUE obj, rb_random_t *rnd, unsigned long limit)
 {
-    rb_random_t *rnd = try_get_rnd(obj);
     if (!rnd) {
 	extern int rb_num_negative_p(VALUE);
 	VALUE lim = ulong_to_num_plus_1(limit);
@@ -864,6 +986,32 @@ rb_random_ulong_limited(VALUE obj, unsigned long limit)
     return limited_rand(&rnd->mt, limit);
 }
 
+static VALUE
+random_ulong_limited_big(VALUE obj, rb_random_t *rnd, VALUE vmax)
+{
+    if (!rnd) {
+	extern int rb_num_negative_p(VALUE);
+	VALUE lim = rb_big_plus(vmax, INT2FIX(1));
+	VALUE v = rb_to_int(rb_funcall2(obj, id_rand, 1, &lim));
+	if (rb_num_negative_p(v)) {
+	    rb_raise(rb_eRangeError, "random number too small %"PRIsVALUE, v);
+	}
+	if (FIX2LONG(rb_big_cmp(vmax, v)) < 0) {
+	    rb_raise(rb_eRangeError, "random number too big %"PRIsVALUE, v);
+	}
+	return v;
+    }
+    return limited_big_rand(&rnd->mt, vmax);
+}
+
+unsigned long
+rb_random_ulong_limited(VALUE obj, unsigned long limit)
+{
+    return random_ulong_limited(obj, try_get_rnd(obj), limit);
+}
+
+static VALUE genrand_bytes(rb_random_t *rnd, long n);
+
 /*
  * call-seq: prng.bytes(size) -> a_string
  *
@@ -875,21 +1023,16 @@ rb_random_ulong_limited(VALUE obj, unsigned long limit)
 static VALUE
 random_bytes(VALUE obj, VALUE len)
 {
-    return rb_random_bytes(obj, NUM2LONG(rb_to_int(len)));
+    return genrand_bytes(get_rnd(obj), NUM2LONG(rb_to_int(len)));
 }
 
-VALUE
-rb_random_bytes(VALUE obj, long n)
+static VALUE
+genrand_bytes(rb_random_t *rnd, long n)
 {
-    rb_random_t *rnd = try_get_rnd(obj);
     VALUE bytes;
     char *ptr;
     unsigned int r, i;
 
-    if (!rnd) {
-	VALUE len = LONG2NUM(n);
-	return rb_funcall2(obj, id_bytes, 1, &len);
-    }
     bytes = rb_str_new(0, n);
     ptr = RSTRING_PTR(bytes);
     for (; n >= SIZEOF_INT32; n -= SIZEOF_INT32) {
@@ -910,6 +1053,17 @@ rb_random_bytes(VALUE obj, long n)
     return bytes;
 }
 
+VALUE
+rb_random_bytes(VALUE obj, long n)
+{
+    rb_random_t *rnd = try_get_rnd(obj);
+    if (!rnd) {
+	VALUE len = LONG2NUM(n);
+	return rb_funcall2(obj, id_bytes, 1, &len);
+    }
+    return genrand_bytes(rnd, n);
+}
+
 static VALUE
 range_values(VALUE vmax, VALUE *begp, VALUE *endp, int *exclp)
 {
@@ -924,55 +1078,81 @@ range_values(VALUE vmax, VALUE *begp, VALUE *endp, int *exclp)
 }
 
 static VALUE
-rand_int(struct MT *mt, VALUE vmax, int restrictive)
+rand_int(VALUE obj, rb_random_t *rnd, VALUE vmax, int restrictive)
 {
     /* mt must be initialized */
-    long max;
     unsigned long r;
 
     if (FIXNUM_P(vmax)) {
-	max = FIX2LONG(vmax);
+	long max = FIX2LONG(vmax);
 	if (!max) return Qnil;
 	if (max < 0) {
 	    if (restrictive) return Qnil;
 	    max = -max;
 	}
-	r = limited_rand(mt, (unsigned long)max - 1);
+	r = random_ulong_limited(obj, rnd, (unsigned long)max - 1);
 	return ULONG2NUM(r);
     }
     else {
 	VALUE ret;
 	if (rb_bigzero_p(vmax)) return Qnil;
-	if (!RBIGNUM_SIGN(vmax)) {
+	if (!BIGNUM_SIGN(vmax)) {
 	    if (restrictive) return Qnil;
             vmax = rb_big_uminus(vmax);
 	}
 	vmax = rb_big_minus(vmax, INT2FIX(1));
 	if (FIXNUM_P(vmax)) {
-	    max = FIX2LONG(vmax);
+	    long max = FIX2LONG(vmax);
 	    if (max == -1) return Qnil;
-	    r = limited_rand(mt, max);
+	    r = random_ulong_limited(obj, rnd, max);
 	    return LONG2NUM(r);
 	}
-	ret = limited_big_rand(mt, vmax);
+	ret = random_ulong_limited_big(obj, rnd, vmax);
 	RB_GC_GUARD(vmax);
 	return ret;
     }
 }
 
+NORETURN(static void domain_error(void));
+static void
+domain_error(void)
+{
+    VALUE error = INT2FIX(EDOM);
+    rb_exc_raise(rb_class_new_instance(1, &error, rb_eSystemCallError));
+}
+
+NORETURN(static void invalid_argument(VALUE));
+static void
+invalid_argument(VALUE arg0)
+{
+    rb_raise(rb_eArgError, "invalid argument - %"PRIsVALUE, arg0);
+}
+
+static VALUE
+check_random_number(VALUE v, const VALUE *argv)
+{
+    switch (v) {
+      case Qfalse:
+	(void)NUM2LONG(argv[0]);
+	break;
+      case Qnil:
+	invalid_argument(argv[0]);
+    }
+    return v;
+}
+
 static inline double
 float_value(VALUE v)
 {
     double x = RFLOAT_VALUE(v);
     if (isinf(x) || isnan(x)) {
-	VALUE error = INT2FIX(EDOM);
-	rb_exc_raise(rb_class_new_instance(1, &error, rb_eSystemCallError));
+	domain_error();
     }
     return x;
 }
 
 static inline VALUE
-rand_range(struct MT* mt, VALUE range)
+rand_range(VALUE obj, rb_random_t* rnd, VALUE range)
 {
     VALUE beg = Qundef, end = Qundef, vmax, v;
     int excl = 0;
@@ -986,17 +1166,17 @@ rand_range(struct MT* mt, VALUE range)
 	if (FIXNUM_P(vmax)) {
 	  fixnum:
 	    if ((max = FIX2LONG(vmax) - excl) >= 0) {
-		unsigned long r = limited_rand(mt, (unsigned long)max);
+		unsigned long r = random_ulong_limited(obj, rnd, (unsigned long)max);
 		v = ULONG2NUM(r);
 	    }
 	}
-	else if (BUILTIN_TYPE(vmax) == T_BIGNUM && RBIGNUM_SIGN(vmax) && !rb_bigzero_p(vmax)) {
+	else if (BUILTIN_TYPE(vmax) == T_BIGNUM && BIGNUM_SIGN(vmax) && !rb_bigzero_p(vmax)) {
 	    vmax = excl ? rb_big_minus(vmax, INT2FIX(1)) : rb_big_norm(vmax);
 	    if (FIXNUM_P(vmax)) {
 		excl = 0;
 		goto fixnum;
 	    }
-	    v = limited_big_rand(mt, vmax);
+	    v = random_ulong_limited_big(obj, rnd, vmax);
 	}
     }
     else if (v = rb_check_to_float(vmax), !NIL_P(v)) {
@@ -1009,17 +1189,12 @@ rand_range(struct MT* mt, VALUE range)
 	    mid = max + min;
 	    max -= min;
 	}
-	else {
-	    float_value(v);
+	else if (isnan(max)) {
+	    domain_error();
 	}
 	v = Qnil;
 	if (max > 0.0) {
-	    if (excl) {
-		r = genrand_real(mt);
-	    }
-	    else {
-		r = genrand_real2(mt);
-	    }
+	    r = random_real(obj, rnd, excl);
 	    if (scale > 1) {
 		return rb_float_new(+(+(+(r - 0.5) * max) * scale) + mid);
 	    }
@@ -1052,7 +1227,7 @@ rand_range(struct MT* mt, VALUE range)
     return v;
 }
 
-static VALUE rand_random(int argc, VALUE *argv, rb_random_t *rnd);
+static VALUE rand_random(int argc, VALUE *argv, VALUE obj, rb_random_t *rnd);
 
 /*
  * call-seq:
@@ -1086,16 +1261,19 @@ static VALUE rand_random(int argc, VALUE *argv, rb_random_t *rnd);
 static VALUE
 random_rand(int argc, VALUE *argv, VALUE obj)
 {
-    return rand_random(argc, argv, get_rnd(obj));
+    VALUE v = rand_random(argc, argv, obj, get_rnd(obj));
+    check_random_number(v, argv);
+    return v;
 }
 
 static VALUE
-rand_random(int argc, VALUE *argv, rb_random_t *rnd)
+rand_random(int argc, VALUE *argv, VALUE obj, rb_random_t *rnd)
 {
     VALUE vmax, v;
+    double max = 0.0;
 
     if (argc == 0) {
-	return rb_float_new(genrand_real(&rnd->mt));
+	goto float_rand;
     }
     else {
 	rb_check_arity(argc, 0, 1);
@@ -1105,28 +1283,41 @@ rand_random(int argc, VALUE *argv, rb_random_t *rnd)
 	v = Qnil;
     }
     else if (!RB_TYPE_P(vmax, T_FLOAT) && (v = rb_check_to_integer(vmax, "to_int"), !NIL_P(v))) {
-	v = rand_int(&rnd->mt, v, 1);
+	v = rand_int(obj, rnd, v, 1);
     }
     else if (v = rb_check_to_float(vmax), !NIL_P(v)) {
-	double max = float_value(v);
-	if (max > 0.0)
-	    v = rb_float_new(max * genrand_real(&rnd->mt));
-	else
+	max = float_value(v);
+	if (max < 0.0) {
 	    v = Qnil;
+	}
+	else {
+	    uint32_t a, b;
+	    double r;
+
+	  float_rand:
+	    a = random_int32(obj, rnd);
+	    b = random_int32(obj, rnd);
+	    r = int_pair_to_real_exclusive(a, b);
+	    if (max > 0.0) r *= max;;
+	    v = rb_float_new(r);
+	}
     }
-    else if ((v = rand_range(&rnd->mt, vmax)) != Qfalse) {
+    else if ((v = rand_range(obj, rnd, vmax)) != Qfalse) {
 	/* nothing to do */
     }
     else {
-	v = Qnil;
-	(void)NUM2LONG(vmax);
-    }
-    if (NIL_P(v)) {
-	VALUE mesg = rb_str_new_cstr("invalid argument - ");
-	rb_str_append(mesg, rb_obj_as_string(argv[0]));
-	rb_exc_raise(rb_exc_new3(rb_eArgError, mesg));
+	return Qfalse;
     }
+    return v;
+}
 
+static VALUE
+rand_random_number(int argc, VALUE *argv, VALUE obj)
+{
+    rb_random_t *rnd = try_get_rnd(obj);
+    VALUE v = rand_random(argc, argv, obj, rnd);
+    if (NIL_P(v)) v = rand_random(0, 0, obj, rnd);
+    else if (!v) invalid_argument(argv[0]);
     return v;
 }
 
@@ -1203,18 +1394,18 @@ static VALUE
 rb_f_rand(int argc, VALUE *argv, VALUE obj)
 {
     VALUE v, vmax, r;
-    struct MT *mt = default_mt();
+    rb_random_t *rnd = rand_start(&default_rand);
 
     if (argc == 0) goto zero_arg;
     rb_scan_args(argc, argv, "01", &vmax);
     if (NIL_P(vmax)) goto zero_arg;
-    if ((v = rand_range(mt, vmax)) != Qfalse) {
+    if ((v = rand_range(Qnil, rnd, vmax)) != Qfalse) {
 	return v;
     }
     vmax = rb_to_int(vmax);
-    if (vmax == INT2FIX(0) || NIL_P(r = rand_int(mt, vmax, 0))) {
+    if (vmax == INT2FIX(0) || NIL_P(r = rand_int(Qnil, rnd, vmax, 0))) {
       zero_arg:
-	return DBL2NUM(genrand_real(mt));
+	return DBL2NUM(genrand_real(&rnd->mt));
     }
     return r;
 }
@@ -1230,7 +1421,9 @@ rb_f_rand(int argc, VALUE *argv, VALUE obj)
 static VALUE
 random_s_rand(int argc, VALUE *argv, VALUE obj)
 {
-    return rand_random(argc, argv, rand_start(&default_rand));
+    VALUE v = rand_random(argc, argv, Qnil, rand_start(&default_rand));
+    check_random_number(v, argv);
+    return v;
 }
 
 #define SIP_HASH_STREAMING 0
@@ -1357,7 +1550,7 @@ rb_reset_random_seed(void)
  */
 
 void
-Init_Random(void)
+InitVM_Random(void)
 {
     Init_RandomSeed2();
     rb_define_global_function("srand", rb_f_srand, -1);
@@ -1386,9 +1579,23 @@ Init_Random(void)
     rb_define_singleton_method(rb_cRandom, "srand", rb_f_srand, -1);
     rb_define_singleton_method(rb_cRandom, "rand", random_s_rand, -1);
     rb_define_singleton_method(rb_cRandom, "new_seed", random_seed, 0);
+    rb_define_singleton_method(rb_cRandom, "raw_seed", random_raw_seed, 1);
     rb_define_private_method(CLASS_OF(rb_cRandom), "state", random_s_state, 0);
     rb_define_private_method(CLASS_OF(rb_cRandom), "left", random_s_left, 0);
 
+    {
+	VALUE m = rb_define_module_under(rb_cRandom, "Formatter");
+	rb_include_module(rb_cRandom, m);
+	rb_define_method(m, "random_number", rand_random_number, -1);
+    }
+}
+
+#undef rb_intern
+void
+Init_Random(void)
+{
     id_rand = rb_intern("rand");
     id_bytes = rb_intern("bytes");
+
+    InitVM(Random);
 }
