@@ -1,3 +1,4 @@
+# -*- coding: us-ascii -*-
 require 'test/unit'
 require 'timeout'
 require 'socket'
@@ -9,7 +10,7 @@ end
 class TestIOWait < Test::Unit::TestCase
 
   def setup
-    if /mswin/ =~ RUBY_PLATFORM
+    if /mswin|mingw/ =~ RUBY_PLATFORM
       @r, @w = Socket.pair(Socket::AF_INET, Socket::SOCK_STREAM, 0)
     else
       @r, @w = IO.pipe
@@ -30,7 +31,7 @@ class TestIOWait < Test::Unit::TestCase
 
   def test_nread_buffered
     @w.syswrite ".\n!"
-    assert_equal ".\n", @r.read(2)
+    assert_equal ".\n", @r.gets
     assert_equal 1, @r.nread
   end
 
@@ -61,12 +62,55 @@ class TestIOWait < Test::Unit::TestCase
   end
 
   def test_wait_forever
-    Thread.new { sleep 0.01; @w.syswrite "." }
+    th = Thread.new { sleep 0.01; @w.syswrite "." }
     assert_equal @r, @r.wait
+  ensure
+    th.join
   end
 
   def test_wait_eof
-    Thread.new { sleep 0.01; @w.close }
-    assert_nil @r.wait
+    th = Thread.new { sleep 0.01; @w.close }
+    ret = nil
+    assert_nothing_raised(Timeout::Error) do
+      Timeout.timeout(0.1) { ret = @r.wait }
+    end
+    assert_equal @r, ret
+  ensure
+    th.join
+  end
+
+  def test_wait_writable
+    assert_equal @w, @w.wait_writable
+  end
+
+  def test_wait_writable_timeout
+    assert_equal @w, @w.wait_writable(0.01)
+    written = fill_pipe
+    assert_nil @w.wait_writable(0.01)
+    @r.read(written)
+    assert_equal @w, @w.wait_writable(0.01)
+  end
+
+  def test_wait_writable_EPIPE
+    fill_pipe
+    @r.close
+    assert_equal @w, @w.wait_writable
+  end
+
+  def test_wait_writable_closed
+    @w.close
+    assert_raise(IOError) { @w.wait_writable }
+  end
+
+private
+
+  def fill_pipe
+    written = 0
+    buf = " " * 4096
+    begin
+      written += @w.write_nonblock(buf)
+    rescue Errno::EAGAIN, Errno::EWOULDBLOCK
+      return written
+    end while true
   end
 end if IO.method_defined?(:wait)
