@@ -3,6 +3,8 @@
 # avoid warnings with -d.
 $install_name ||= nil
 $so_name ||= nil
+arch = $arch or raise "missing -arch"
+version = $version or raise "missing -version"
 
 srcdir = File.expand_path('../..', __FILE__)
 $:.replace [srcdir+"/lib"] unless defined?(CROSS_COMPILING)
@@ -16,7 +18,6 @@ unless File.directory?(dir = File.dirname(rbconfig_rb))
   FileUtils.makedirs(dir, :verbose => true)
 end
 
-version = RUBY_VERSION
 config = ""
 def config.write(arg)
   concat(arg.to_s)
@@ -24,17 +25,7 @@ end
 $stdout = config
 
 fast = {'prefix'=>TRUE, 'ruby_install_name'=>TRUE, 'INSTALL'=>TRUE, 'EXEEXT'=>TRUE}
-print %[
-# This file was created by #{mkconfig} when ruby was built.  Any
-# changes made to this file will be lost the next time ruby is built.
 
-module RbConfig
-  RUBY_VERSION == "#{version}" or
-    raise "ruby lib version (#{version}) doesn't match executable version (\#{RUBY_VERSION})"
-
-]
-
-arch = RUBY_PLATFORM
 win32 = /mswin/ =~ arch
 universal = /universal.*darwin/ =~ arch
 v_fast = []
@@ -42,6 +33,8 @@ v_others = []
 vars = {}
 continued_name = nil
 continued_line = nil
+install_name = nil
+so_name = nil
 File.foreach "config.status" do |line|
   next if /^#/ =~ line
   name = nil
@@ -61,6 +54,7 @@ File.foreach "config.status" do |line|
     next if !continued_line
     continued_line << $1
     next if $2
+    continued_line.each {|s| s.sub!(/\\n\z/, "\n")}
     val = continued_line.join
     name = continued_name
     continued_line = nil
@@ -71,12 +65,16 @@ File.foreach "config.status" do |line|
   if name
     case name
     when /^(?:ac_.*|configure_input|(?:top_)?srcdir|\w+OBJS)$/; next
-    when /^(?:X|(?:MINI|RUN)RUBY$)/; next
-    when /^(?:MAJOR|MINOR|TEENY)$/; next
-    when /^RUBY_INSTALL_NAME$/; next if $install_name
-    when /^RUBY_SO_NAME$/; next if $so_name
+    when /^(?:X|(?:MINI|RUN|(?:HAVE_)?BASE|BOOTSTRAP|BTEST)RUBY(?:_COMMAND)?$)/; next
+    when /^INSTALLDOC|TARGET$/; next
+    when /^DTRACE/; next
+    when /^(?:MAJOR|MINOR|TEENY)$/; vars[name] = val; next
+    when /^LIBRUBY_D?LD/; next
+    when /^RUBY_INSTALL_NAME$/; next vars[name] = (install_name = val).dup if $install_name
+    when /^RUBY_SO_NAME$/; next vars[name] = (so_name = val).dup if $so_name
     when /^arch$/; if val.empty? then val = arch else arch = val end
-    when /^sitearch/; val = '$(arch)' if val.empty?
+    when /^sitearch$/; val = '$(arch)' if val.empty?
+    when /^DESTDIR$/; next
     end
     case val
     when /^\$\(ac_\w+\)$/; next
@@ -111,8 +109,9 @@ File.foreach "config.status" do |line|
         end
       end
     end
+    eq = win32 && vars[name] ? '<< "\n"' : '='
+    vars[name] = val
     if name == "configure_args"
-      val.gsub!(/ +(?!-)/, "=") if win32
       val.gsub!(/--with-out-ext/, "--without-ext")
     end
     val = val.gsub(/\$(?:\$|\{?(\w+)\}?)/) {$1 ? "$(#{$1})" : $&}.dump
@@ -127,16 +126,17 @@ File.foreach "config.status" do |line|
       if universal
         val.sub!(/universal/, %q[#{arch && universal[/(?:\A|\s)#{Regexp.quote(arch)}=(\S+)/, 1] || '\&'}])
       end
+    when /^includedir$/
+      val = '"$(SDKROOT)"'+val if /darwin/ =~ arch
     end
-    v = "  CONFIG[\"#{name}\"] #{win32 && vars[name] ? '<< "\n"' : '='} #{val}\n"
-    vars[name] = true
+    v = "  CONFIG[\"#{name}\"] #{eq} #{val}\n"
     if fast[name]
       v_fast << v
     else
       v_others << v
     end
     case name
-    when "ruby_version"
+    when "RUBY_PROGRAM_VERSION"
       version = val[/\A"(.*)"\z/, 1]
     end
   end
@@ -145,8 +145,37 @@ end
 
 drive = File::PATH_SEPARATOR == ';'
 
-prefix = "/lib/ruby/#{version}/#{arch}"
-print "  TOPDIR = File.dirname(__FILE__).chomp!(#{prefix.dump})\n"
+def vars.expand(val, config = self)
+  newval = val.gsub(/\$\$|\$\(([^()]+)\)|\$\{([^{}]+)\}/) {
+    var = $&
+    if !(v = $1 || $2)
+      '$'
+    elsif key = config[v = v[/\A[^:]+(?=(?::(.*?)=(.*))?\z)/]]
+      pat, sub = $1, $2
+      config[v] = false
+      config[v] = expand(key, config)
+      key = key.gsub(/#{Regexp.quote(pat)}(?=\s|\z)/n) {sub} if pat
+      key
+    else
+      var
+    end
+  }
+  val.replace(newval) unless newval == val
+  val
+end
+prefix = vars.expand(vars["prefix"] ||= "")
+rubyarchdir = vars.expand(vars["rubyarchdir"] ||= "")
+relative_archdir = rubyarchdir.rindex(prefix, 0) ? rubyarchdir[prefix.size..-1] : rubyarchdir
+puts %[\
+# This file was created by #{mkconfig} when ruby was built.  Any
+# changes made to this file will be lost the next time ruby is built.
+
+module RbConfig
+  RUBY_VERSION.start_with?("#{version[/^[0-9]+\.[0-9]+\./] || version}") or
+    raise "ruby lib version (#{version}) doesn't match executable version (\#{RUBY_VERSION})"
+
+]
+print "  TOPDIR = File.dirname(__FILE__).chomp!(#{relative_archdir.dump})\n"
 print "  DESTDIR = ", (drive ? "TOPDIR && TOPDIR[/\\A[a-z]:/i] || " : ""), "'' unless defined? DESTDIR\n"
 print <<'ARCH' if universal
   arch_flag = ENV['ARCHFLAGS'] || ((e = ENV['RC_ARCHS']) && e.split.uniq.map {|a| "-arch #{a}"}.join(' '))
@@ -175,32 +204,44 @@ end
   print "  CONFIG[#{v.dump}] = #{versions[v].dump}\n"
 end
 
-dest = drive ? /= \"(?!\$[\(\{])(?:[a-z]:)?/i : /= \"(?!\$[\(\{])/
+dest = drive ? %r'= "(?!\$[\(\{])(?i:[a-z]:)' : %r'= "(?!\$[\(\{])'
+v_disabled = {}
 v_others.collect! do |x|
-  if /^\s*CONFIG\["(?!abs_|old)[a-z]+(?:_prefix|dir)"\]/ === x
+  if /^\s*CONFIG\["((?!abs_|old)[a-z]+(?:_prefix|dir))"\]/ === x
+    name = $1
+    if /= "no"$/ =~ x
+      v_disabled[name] = true
+      v_others.delete(name)
+      next
+    end
     x.sub(dest, '= "$(DESTDIR)')
   else
     x
   end
 end
+v_others.compact!
 
 if $install_name
+  if install_name and vars.expand("$(RUBY_INSTALL_NAME)") == $install_name
+    $install_name = install_name
+  end
   v_fast << "  CONFIG[\"ruby_install_name\"] = \"" + $install_name + "\"\n"
   v_fast << "  CONFIG[\"RUBY_INSTALL_NAME\"] = \"" + $install_name + "\"\n"
 end
 if $so_name
+  if so_name and vars.expand("$(RUBY_SO_NAME)") == $so_name
+    $so_name = so_name
+  end
   v_fast << "  CONFIG[\"RUBY_SO_NAME\"] = \"" + $so_name + "\"\n"
 end
 
 print(*v_fast)
 print(*v_others)
+print <<EOS if /darwin/ =~ arch
+  CONFIG["SDKROOT"] = ENV["SDKROOT"] || "" # don't run xcrun everytime, usually useless.
+EOS
 print <<EOS
-  CONFIG["rubylibdir"] = "$(rubylibprefix)/$(ruby_version)"
-  CONFIG["archdir"] = "$(rubylibdir)/$(arch)"
-  CONFIG["sitelibdir"] = "$(sitedir)/$(ruby_version)"
-  CONFIG["sitearchdir"] = "$(sitelibdir)/$(sitearch)"
-  CONFIG["vendorlibdir"] = "$(vendordir)/$(ruby_version)"
-  CONFIG["vendorarchdir"] = "$(vendorlibdir)/$(sitearch)"
+  CONFIG["archdir"] = "$(rubyarchdir)"
   CONFIG["topdir"] = File.dirname(__FILE__)
   MAKEFILE_CONFIG = {}
   CONFIG.each{|k,v| MAKEFILE_CONFIG[k] = v.dup}
@@ -234,7 +275,6 @@ print <<EOS
     )
   end
 end
-autoload :Config, "rbconfig/obsolete.rb" # compatibility for ruby-1.8.4 and older.
 CROSS_COMPILING = nil unless defined? CROSS_COMPILING
 EOS
 
