@@ -1,10 +1,10 @@
-#define RUBY_VERSION "1.9.3"
+#define RUBY_VERSION "2.3.0"
+#define RUBY_RELEASE_DATE "2015-11-11"
 #define RUBY_PATCHLEVEL -1
 
-#define RUBY_RELEASE_DATE "2011-07-31"
-#define RUBY_RELEASE_YEAR 2011
-#define RUBY_RELEASE_MONTH 7
-#define RUBY_RELEASE_DAY 31
+#define RUBY_RELEASE_YEAR 2015
+#define RUBY_RELEASE_MONTH 11
+#define RUBY_RELEASE_DAY 11
 
 #include "ruby/version.h"
 
@@ -17,7 +17,7 @@
 #endif
 
 #if RUBY_PATCHLEVEL == -1
-#define RUBY_PATCHLEVEL_STR "dev"
+#define RUBY_PATCHLEVEL_STR "preview1"
 #else
 #define RUBY_PATCHLEVEL_STR "p"STRINGIZE(RUBY_PATCHLEVEL)
 #endif
@@ -25,12 +25,15 @@
 #ifndef RUBY_REVISION
 # include "revision.h"
 #endif
-# ifndef RUBY_REVISION
+#ifndef RUBY_REVISION
 # define RUBY_REVISION 0
 #endif
 
 #if RUBY_REVISION
-# ifdef RUBY_BRANCH_NAME
+# if RUBY_PATCHLEVEL == -1
+#  ifndef RUBY_BRANCH_NAME
+#   define RUBY_BRANCH_NAME "trunk"
+#  endif
 #  define RUBY_REVISION_STR " "RUBY_BRANCH_NAME" "STRINGIZE(RUBY_REVISION)
 # else
 #  define RUBY_REVISION_STR " revision "STRINGIZE(RUBY_REVISION)
@@ -41,11 +44,11 @@
 
 # define RUBY_DESCRIPTION	    \
     "ruby "RUBY_VERSION		    \
-    RUBY_PATCHLEVEL_STR             \
+    RUBY_PATCHLEVEL_STR		    \
     " ("RUBY_RELEASE_DATE	    \
     RUBY_REVISION_STR") "	    \
     "["RUBY_PLATFORM"]"
-# define RUBY_COPYRIGHT 	    \
+# define RUBY_COPYRIGHT		    \
     "ruby - Copyright (C) "	    \
     STRINGIZE(RUBY_BIRTH_YEAR)"-"   \
     STRINGIZE(RUBY_RELEASE_YEAR)" " \
